import { __decorate, __param } from 'vscode/external/tslib/tslib.es6.js';
import { Disposable } from 'vscode/vscode/vs/base/common/lifecycle';
import Severity$1 from 'vscode/vscode/vs/base/common/severity';
import { getCodeEditor } from 'vscode/vscode/vs/editor/browser/editorBrowser';
import { IEditorService } from 'vscode/vscode/vs/workbench/services/editor/common/editorService.service';
import { ILanguageStatusService } from 'vscode/vscode/vs/workbench/services/languageStatus/common/languageStatusService.service';
import { Registry } from 'vscode/vscode/vs/platform/registry/common/platform';
import { Extensions } from 'vscode/vscode/vs/workbench/common/contributions';
import { localizeWithPath } from 'vscode/vscode/vs/nls';
import { FoldingController } from 'vscode/vscode/vs/editor/contrib/folding/browser/folding';
import { ColorDetector } from 'vscode/vscode/vs/editor/contrib/colorPicker/browser/colorDetector';

const _moduleId = "vs/workbench/contrib/limitIndicator/browser/limitIndicator.contribution";
const openSettingsCommand = 'workbench.action.openSettings';
const configureSettingsLabel = ( localizeWithPath(_moduleId, 0, "Configure"));
let LimitIndicatorContribution = class LimitIndicatorContribution extends Disposable {
    constructor(editorService, languageStatusService) {
        super();
        const accessors = [( (new ColorDecorationAccessor())), ( (new FoldingRangeAccessor()))];
        const statusEntries = ( (accessors.map(
            indicator => ( (new LanguageStatusEntry(languageStatusService, indicator)))
        )));
        statusEntries.forEach(entry => this._register(entry));
        let control;
        const onActiveEditorChanged = () => {
            const activeControl = editorService.activeTextEditorControl;
            if (activeControl === control) {
                return;
            }
            control = activeControl;
            const editor = getCodeEditor(activeControl);
            statusEntries.forEach(statusEntry => statusEntry.onActiveEditorChanged(editor));
        };
        this._register(editorService.onDidActiveEditorChange(onActiveEditorChanged));
        onActiveEditorChanged();
    }
};
LimitIndicatorContribution = ( (__decorate([
    ( (__param(0, IEditorService))),
    ( (__param(1, ILanguageStatusService)))
], LimitIndicatorContribution)));
class ColorDecorationAccessor {
    constructor() {
        this.id = 'decoratorsLimitInfo';
        this.name = ( localizeWithPath(_moduleId, 1, 'Color Decorator Status'));
        this.label = ( localizeWithPath(_moduleId, 2, 'Color Decorators'));
        this.source = ( localizeWithPath(_moduleId, 3, 'Color Decorators'));
        this.settingsId = 'editor.colorDecoratorsLimit';
    }
    getLimitReporter(editor) {
        return ColorDetector.get(editor)?.limitReporter;
    }
}
class FoldingRangeAccessor {
    constructor() {
        this.id = 'foldingLimitInfo';
        this.name = ( localizeWithPath(_moduleId, 4, 'Folding Status'));
        this.label = ( localizeWithPath(_moduleId, 5, 'Folding Ranges'));
        this.source = ( localizeWithPath(_moduleId, 6, 'Folding'));
        this.settingsId = 'editor.foldingMaximumRegions';
    }
    getLimitReporter(editor) {
        return FoldingController.get(editor)?.limitReporter;
    }
}
class LanguageStatusEntry {
    constructor(languageStatusService, accessor) {
        this.languageStatusService = languageStatusService;
        this.accessor = accessor;
    }
    onActiveEditorChanged(editor) {
        if (this._indicatorChangeListener) {
            this._indicatorChangeListener.dispose();
            this._indicatorChangeListener = undefined;
        }
        let info;
        if (editor) {
            info = this.accessor.getLimitReporter(editor);
        }
        this.updateStatusItem(info);
        if (info) {
            this._indicatorChangeListener = info.onDidChange(_ => {
                this.updateStatusItem(info);
            });
            return true;
        }
        return false;
    }
    updateStatusItem(info) {
        if (this._limitStatusItem) {
            this._limitStatusItem.dispose();
            this._limitStatusItem = undefined;
        }
        if (info && info.limited !== false) {
            const status = {
                id: this.accessor.id,
                selector: '*',
                name: this.accessor.name,
                severity: Severity$1.Warning,
                label: this.accessor.label,
                detail: ( localizeWithPath(_moduleId, 7, 'only {0} shown for performance reasons', info.limited)),
                command: { id: openSettingsCommand, arguments: [this.accessor.settingsId], title: configureSettingsLabel },
                accessibilityInfo: undefined,
                source: this.accessor.source,
                busy: false
            };
            this._limitStatusItem = this.languageStatusService.addStatus(status);
        }
    }
    dispose() {
        this._limitStatusItem = undefined;
        this._indicatorChangeListener = undefined;
    }
}
( (Registry.as(Extensions.Workbench))).registerWorkbenchContribution(LimitIndicatorContribution, 3 );

export { LimitIndicatorContribution };
